/*
 * This file contains the functions for a termstack.
 *
 * A termstack is a container which can have up to MAXSUBSTACK clients.
 * Events are sent successively to all clients. Clients see all events
 * generated by tjose earlier in the queue (NO feedback).
 *
 * A term understands the following commands :
 *
 * Command Short Parameters
 * attach  +     stack_to_attach
 * detach  -     stack_to_detach
 */

#include "evstack.h"
#include "parsearg.h"

#define MAXSUBSTACK 16

typedef struct termstack_state
{
	int refcount;
	evstack *substacks[MAXSUBSTACK];
} termstack_state;

static termstack_state defaultstate = {
	0,
	{ NULL, }
};

#define STACKTYPEDEF termstack_state
#include "substack.inc"

#define CLASSNAME "termstack"

static ev_id termstack_id;
static int termstack_use=0;
static int termstack_open(evstack *blank);
static int termstack_close(evstack *slf);
static void termstack_handler(evstack *slf,ggi_event *ev,int *left);
static int termstack_uses(evstack *slf);
static int termstack_write(evstack *slf, char *str);
static int termstack_read(evstack *slf, int lineno, int maxlen, char *str);

static evstack_operations termstack_ops = {
	termstack_open,
	termstack_close,
	termstack_handler,
	termstack_uses,
	termstack_write,
	termstack_read
};

static int termstack_open(evstack *blank)
{
	termstack_state *dat;

	sprintf(blank->name,"TermStack v1.0 (instance %.4x)",
		blank->id & EV_ID_MASK);

	blank->maintainer="becka@sunserver1.rz.uni-duesseldorf.de";

	if ((dat=blank->private))
	{
		dat->refcount++;
		DEBUG2("Stack cloned (rc %d).",dat->refcount);
	}
	else
	{	
		blank->private=dat=(termstack_state *)ev_malloc(sizeof(termstack_state));
		memcpy(dat,&defaultstate,sizeof(termstack_state));
		DEBUG2("Stack created (rc %d).",dat->refcount);
	}

        MOD_INC_USE_COUNT;
        termstack_use++;

	return 0;
}

static int termstack_close(evstack *slf)
{
	termstack_state *des=slf->private;
        
	if (--des->refcount==0)		/* No more instances left */
	{
		ev_free(des);
		DEBUG2("Stack destroyed.",0);
	}

        MOD_DEC_USE_COUNT;
	termstack_use--;
	
	return 0;
}

/* Event handler code.
 *
 * Stack-behaviour. Members will be fed in order earlier ones not seeing
 * events generated by their successors.
 */

static void termstack_handler(evstack *slf,ggi_event *ev,int *left)
{ 
	int x;
	ggi_cmd_event *evdat;
	struct termstack_state *ss=slf->private;

	reprocess:
	/* send all to the attached substacks. */
	for(x=0;x<MAXSUBSTACK&&ss->substacks[x];x++)
	{
  	  	DEBUG1("%s sending to %s\n",slf->name,ss->substacks[x]->name);
		ev_send(ev,left,NULL,ss->substacks[x]);
	}

	x=0;
	/* now check for a reprocess-request */
	for(;ev->size;ev=ev_evpage_next(ev))
	{
		switch(ev->any.type)
		{ 
		  case evCommand:
		  	evdat=(ggi_cmd_event *)ev;
		  	switch(evdat->code)
		  	{
		  	  case CmdStkReprocess:
		  	  	ev_consume(ev);
		  	  	x=1;
				break;
		  	  default:
		  	  	break;
			}
			break;
		  default: 
		  	break;
		}
	}
	if (x==1) goto reprocess;
}

static int termstack_uses(evstack *slf)
{
	return termstack_use;
}

static int termstack_write(evstack *slf, char *str)
{
	int rc,help;
	char buffer[256];
	struct termstack_state *ss=slf->private;
	
	DEBUG1("%s->write:%s",slf->name,str);

	if ( 1 != sscanf(str,"%s%n",buffer,&help) ) return 0;	/* empty */
	str+=help;
	DEBUG1("%s->write: command=%s,args=%s",slf->name,buffer,str);
	
	if (strcmp(buffer,"attach")==0 ||
	    strcmp(buffer,"+")==0 ) 
	{
		evstack *hlpstk;
		if ( 1 != sscanf(str,"%s%n",buffer,&help) ) 
		{
			DEBUG1("Invalid attach command.\n",0);return -EINVAL;
		}
		DEBUG1("attach: args=%s",buffer);
		if (!(hlpstk=par_string_to_stack(buffer))) return -EINVAL;
		DEBUG1("attach: stk=%p",hlpstk);
		rc=attach_ev_stack(slf,hlpstk,NULL);
    		if (rc<0) return rc;

    		hlpstk->report_to=slf;
    		return 0;
	}
	else if (strcmp(buffer,"detach")==0 ||
		 strcmp(buffer,"-")==0 ) 
	{
		evstack *hlpstk;
		if ( 1 != sscanf(str,"%s%n",buffer,&help) ) 
		{
			DEBUG1("Invalid detach command.\n",0);return -EINVAL;
		}
		if (!(hlpstk=par_string_to_stack(buffer))) return -EINVAL;
		rc=detach_ev_stack(slf,hlpstk);
    		if (rc<0) return rc;

    		hlpstk->report_to=NULL;
    		return 0;
	}
	return -EINVAL;
}

static int termstack_read(evstack *slf,int lineno,int maxlen,char *str)
{
	struct termstack_state *ss=slf->private;
	if (lineno==0) 
	{
		sprintf(str,"# TermStack %s consists of :",par_stack_to_string(slf)); return 0;
	} else
	if (lineno<=MAXSUBSTACK)
	{
		if (ss->substacks[lineno-1])
		{	sprintf(str,"+ %s   #%s",par_stack_to_string(ss->substacks[lineno-1]),
				ss->substacks[lineno-1]->name); 
			return 0;
		}
	}
	return -EINVAL;
}
#ifdef MODULE
#define init_termstack init_module
#endif

int init_termstack(void)
{ 
	termstack_id=ev_register_class(CLASSNAME,&termstack_ops);
	printk("Registered %s.\n",CLASSNAME);
	return 0;
}

#ifdef MODULE
void cleanup_module(void)
{ 
	ev_unregister_class(termstack_id);
	printk("Unregistered %s.\n",CLASSNAME);
}
#endif
