
/* ======================================================================
 * 
 *	te_demo.c
 *
 *	Copyright (c) 1998  Andrew Apted.
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License.
 *
 *	Demonstrates the TrueEmu display target.
 *
 *	USAGE:  te_demo  [ depth [ size [ target ]]]
 *
 *	Depth is 1, 4, 8, 15, 16, 24 or 32.  A mode specifier can be
 *	appended, of the form <letter>[<digit>], where letter is one of 
 *	r, c, p, d for palette mode (d means use default), and digit is
 *	one of 0, 2, 4 for dither mode.  Some examples: 8p, 15d4.
 *
 *	Size is <width>x<height>.  A depth specifier can be appended, of
 *	the for x<depth>.  Some examples: 320x200, 640x480x32.
 *
 * ======================================================================
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <time.h>
#include <sys/time.h>
#include <math.h>

#include <ggi/libggi.h>


#ifndef MAX
#define MAX(a,b)  ((a) > (b) ? (a) : (b))
#endif 

#define GGI_MAX_INTENSITY  0xffff


static void parse_size(char *size, int *w, int *h, int *gt)
{
	char *s;
	int bpp;

	if ((s = strchr(size, 'x')) == NULL) {
		fprintf(stderr, "te_demo: Bad size '%s'.\n", size);
		exit(1);
	}

	*s = ' ';

	if ((s = strchr(size, 'x')) == NULL) {

		if (sscanf(size, "%d %d", w, h) != 2) {
			fprintf(stderr, "te_demo: Bad size '%s'.\n", size);
			exit(1);
		}
		return;
	}

	*s = ' ';

	if (sscanf(size, "%d %d %d", w, h, &bpp) != 3) {
		fprintf(stderr, "te_demo: Bad size '%s'.\n", size);
		exit(1);
	}

	/* convert bpp to graphtype */

	switch (bpp) {

		case 24: *gt = GT_24BIT; break;
		case 32: *gt = GT_32BIT; break;

		default:
			fprintf(stderr, "te_demo: Bad depth '%d'.\n", bpp);
			exit(1);
	}
}


void main(int argc, char **argv, char **envp)
{ 
	ggi_visual_t vis;

	int x, y;
	int w=640, h=480, gt=GT_24BIT;

	int quit=0;

	char *mode = "8";
	char *target = "display-KGI:/dev/graphic";
	char target_buf[200];


	/* handle args */

	if ((argc > 4) ||
	    ((argc > 1) &&
	     ((strcmp(argv[1], "-h") == 0) ||
	      (strcmp(argv[1], "--help") == 0)))) {
	
		fprintf(stderr, "\nUSAGE:  te_demo  [ depth [ size "
			"[ target ]]]\n\n");
		exit(1);
	}

	if (argc > 1) {
		mode = argv[1];
	}
	if (argc > 2) {
		parse_size(argv[2], &w, &h, &gt);
	}
	if (argc > 3) {
		target = argv[3];
	}


	/* setup graphics mode */

	if (ggiInit() != 0) {
		fprintf(stderr, "%s: unable to initialize libggi, exiting.\n",
			argv[0]);
		exit(1);
	}

	sprintf(target_buf, "display-trueemu:%s:%s", mode, target);

	vis = ggiOpen(target_buf);

	if (vis == NULL) {
		fprintf(stderr, "te_demo: Failed to open visual '%s'.\n", 
			target_buf);
		exit(1);
	}

	if (ggiSetGraphMode(vis, w, h, w, h, gt) < 0) {
		fprintf(stderr, "te_demo: Couldn't set graph mode.\n");
		ggiClose(vis);
		exit(1);
	}

	
	/* draw a pattern in 24 bit mode */

	for (y=0; y < h/2; y++)
	for (x=0; x < w/2; x++) {

		int xi = x * GGI_MAX_INTENSITY / (w/2);
		int yi = y * GGI_MAX_INTENSITY / (h/2);
		int zi = MAX(xi, yi);

		int nxi = GGI_MAX_INTENSITY - xi;
		int nyi = GGI_MAX_INTENSITY - yi;


		ggi_color RG_col;
		ggi_color RB_col;
		ggi_color GB_col;
		ggi_color bw_col;

		RG_col.r = nyi;  RG_col.g = nxi;  RG_col.b = 0;
		RB_col.r = nyi;  RB_col.g = 0;    RB_col.b = xi;
		GB_col.r = 0;    GB_col.g = nxi;  GB_col.b = yi;
		bw_col.r = zi;   bw_col.g = zi;   bw_col.b = zi;

		ggiPutPixel(vis, x,     y,     ggiMapColor(vis, & RG_col));
		ggiPutPixel(vis, x+w/2, y,     ggiMapColor(vis, & RB_col));
		ggiPutPixel(vis, x,     y+h/2, ggiMapColor(vis, & GB_col));
		ggiPutPixel(vis, x+w/2, y+h/2, ggiMapColor(vis, & bw_col));
	}

	ggiFlush(vis);


	while (! quit) {

		int sym = ggiKeyRead(vis);

		/* Note: this sym fiddling is a hack */

		if ((KTYP(sym) == 0xf0) || (KTYP(sym) == 0xfb)) {
			sym &= 0x0fff;
		}

		if (KTYP(sym) == 0x0b) {
			sym &= 0x00ff;
		}

		switch (sym) {
		
			case 'q': case 'Q':
			case 'x': case 'X':
			case '\033':
				quit = 1;
				break;

			default:
				if (((0x20 <= sym) && (sym <= 0x7e)) ||
				    ((0xa0 <= sym) && (sym <= 0xff))) {
					printf("Latin1 key '%c'\n", sym);

				} else if (sym <= 0xff) {
					printf("Latin1 key 0x%02x.\n", sym);

				} else {
					printf("Key sym=0x%04x.\n", sym);
				}
		}
	}
		

	ggiClose(vis);
	ggiExit();

	exit(0);
}
