/* 
   Graphics library for GGI. Mode management.

   Copyright (C) 1998 Marcus Sundberg - marcus@ggi-project.org
   Copyright (C) 1997 Jason McMullan - jmcc@cs.cmu.edu

   If you do any modifications, I would like you to send diffs to me
   to allow for collecting a more and more complete set of drivers and
   to improve existing ones.

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public
   License along with this library; if not, write to the Free
   Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
 * This file is included by Xlib/mode.c and X/mode.c
 * It contains the following common functions:
 *
 *   static int _GGIgetbpp(ggi_visual *vis);
 *   static ggi_graphtype _GGIautodepth(ggi_visual *vis);
 *   static int _GGIbasiccheck(ggi_visual *vis,ggi_mode *tm,
                               XVisualInfo *vinfo);
 *   static int _GGImaskgetshift(int mask);
 *   int GGIcheckmode(ggi_visual *vis,ggi_mode *tm);
 *   int GGIgetmode(ggi_visual *vis,ggi_mode *tm);
 */  

static int _GGIgetbpp(ggi_visual *vis)
{
	XImage *bppcheckimage;
	unsigned int bits_per_pixel=0;
	
	if ((bppcheckimage=XGetImage(XLIB_PRIV(vis)->display,
								 RootWindow(XLIB_PRIV(vis)->display,
											XLIB_PRIV(vis)->screen),
								 0, 0, 1, 1, AllPlanes, ZPixmap)) != NULL) {
		bits_per_pixel=bppcheckimage->bits_per_pixel;
		XDestroyImage(bppcheckimage);
	}
	return bits_per_pixel;
}

static ggi_graphtype _GGIautodepth(ggi_visual *vis)
{
	XVisualInfo vinfo;
	XVisualInfo *vinfos;
	int nitems;
	int depth=0;
	ggi_graphtype gt;

	vinfos = XGetVisualInfo(XLIB_PRIV(vis)->display, VisualNoMask,
							&vinfo, &nitems);
	while (nitems--) {
		if (vinfos[nitems].depth > depth)
			depth = vinfos[nitems].depth;
	}
	XFree(vinfos);

	switch (depth) {
	case 1:
		gt = GT_1BIT;
		break;
	case 4:
		gt = GT_4BIT;
		break;
	case 8:
		gt = GT_8BIT;
		break;
	case 15:
		gt = GT_15BIT;
		break;
	case 16:
		gt = GT_16BIT;
		break;
	case 24:
	case 32:
		switch (_GGIgetbpp(vis)) {
		case 24:
			gt = GT_24BIT;
			break;
		case 32:
			gt = GT_32BIT;
			break;
		default:
			gt = GT_LAST;
			break;
		}
		break;
	default:
		gt = GT_LAST;
		break;
	}
	
	return gt;
}

static int _GGIbasiccheck(ggi_visual *vis,ggi_mode *tm, XVisualInfo *vinfo)
{
	struct Xhooks *xhook=VIS_PRIVATE(vis);
	unsigned int bits_per_pixel;

	bits_per_pixel = _GGIgetbpp(vis);

	switch(tm->graphtype) {
	case GT_8BIT:
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 8, PseudoColor, vinfo) )
			return -1;
		if((tm->virt.x & ~3) != tm->virt.x) {
			tm->virt.x=((tm->virt.x+3) & ~3);
			return -1;
		}
		break;
	case GT_15BIT:
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 15, TrueColor, vinfo) )
			return -1;
		if((tm->virt.x & ~1) != tm->virt.x) {
			tm->virt.x=((tm->virt.x+1) & ~1);
			return -1;
		}
		break;
	case GT_16BIT:
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 16, TrueColor, vinfo) )
			return -1;
		if((tm->virt.x & ~1) != tm->virt.x) {
			tm->virt.x=((tm->virt.x+1) & ~1);
			return -1;
		}
		break;
	case GT_24BIT:
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 24, TrueColor, vinfo) || bits_per_pixel != 24 )
			return -1;
		break;
	case GT_32BIT:
		if ( ! XMatchVisualInfo (xhook->display, xhook->screen, 24, TrueColor, vinfo) || bits_per_pixel != 32 )
			return -1;
		break;
	default:
		return -1;
	}

	return 0;
}

static int _GGImaskgetshift(int mask)
{
	int shift;
	for (shift=16; mask!=0; mask=mask>>1) {
		shift--;
	}
	return shift;
}

/**********************************/
/* check any mode (text/graphics) */
/**********************************/
int GGIcheckmode(ggi_visual *vis,ggi_mode *tm)
{
	XVisualInfo vinfo;      /* Dummy here */
	struct Xhooks *xhook;
	int err = 0;

	if (vis==NULL) {
		DPRINT("Visual==NULL\n");
		return -1;
	}

	xhook=VIS_PRIVATE(vis);

	if(tm->visible.x==GGI_AUTO) 
		tm->visible.x = tm->virt.x==GGI_AUTO? 640 : tm->virt.x;
	if(tm->visible.y==GGI_AUTO)
		tm->visible.y = tm->virt.y==GGI_AUTO? 480 : tm->virt.y;
	if(tm->virt.x==GGI_AUTO) {
		tm->virt.x = (tm->visible.x+3) & ~3;
	}
	if(tm->virt.y==GGI_AUTO)
		tm->virt.y = tm->visible.y;

	if(tm->virt.x < tm->visible.x) {
		tm->virt.x = (tm->visible.x+3) & ~3;
		err = -1;
	}
	if(tm->virt.y < tm->visible.y) {
		tm->virt.y = tm->visible.y;
		err = -1;
	}

	if(	(tm->dpp.x!=1 && tm->dpp.x!=GGI_AUTO) ||
		(tm->dpp.y!=1 && tm->dpp.y!=GGI_AUTO)) {
		err = -1;
	}
	tm->dpp.x = tm->dpp.y = 1;
	
	switch(tm->graphtype) {
	case GT_8BIT:
	case GT_15BIT:
	case GT_16BIT:
	case GT_24BIT:
	case GT_32BIT:
		if (_GGIbasiccheck(vis,tm,&vinfo) == 0)
			break;
		else
			err = -1;
	case GGI_AUTO:
	default:	
		if(tm->graphtype!=GGI_AUTO) err = -1;
			/* Get the default depth */
		tm->graphtype=_GGIautodepth(vis);
		break;
	}

	return err;
}

/************************/
/* get the current mode */
/************************/
int GGIgetmode(ggi_visual *vis,ggi_mode *tm)
{
	DPRINT("In GGIgetmode(%p,%p)\n",vis,tm);
	if (vis==NULL)
		return -1;
	
		/* We assume the mode in the visual to be o.k. */
	memcpy(tm,VIS_MODE(vis),sizeof(ggi_mode));

	return 0;
}
